//	<><><><><><><><><><><><><>  Plot3DT.cpp  <><><><><><><><><><><><><><> */
//		
// ----------------------------------------------------
// (C) Copyright Clark Dailey, Inc. 2006
//
// Implementation of Plot3DT C++ class
//
// --------------------------------------------------------------------------
//

#include "stdafx.h"	// precompiled header
#include <stdio.h>
#include <float.h>
#include "Plot3DT.h"
#include "SurfX3D.h"


// ------------------------------------------------------------
//   Point3DT methods
// ------------------------------------------------------------
void Point3DT::Load(unsigned i,unsigned j,double x,double y,double z,double r,double g,double b)
{
	m_i=i;
	m_j=j;
	m_xf=x;
	m_yf=y;
	m_zf=z;
	m_red=r;
	m_green=g;
	m_blue=b;
}


// ------------------------------------------------------------
//     Triangle3DT methods
// ------------------------------------------------------------
int Triangle3DT::IsDefiningPoint(Point3DT* pPnt)
{
	unsigned i,j;
	i = pPnt->m_i;
	j = pPnt->m_j;
	if (m_pPnt[0]->m_i == i && m_pPnt[0]->m_j == j) return(1);
	if (m_pPnt[1]->m_i == i && m_pPnt[1]->m_j == j) return(1);
	if (m_pPnt[2]->m_i == i && m_pPnt[2]->m_j == j) return(1);
	return(0);
}

// ------------------------------------------------------------
// is a given xp,yp point within the triangle
// returns: 0=no, 1=point is within triangle
int Triangle3DT::IsInscribed(double xp,double yp)
{
	double dir;

	// first check bounding rectangle
	if (xp<m_xptmin || xp>m_xptmax || yp<m_yptmin || yp>m_yptmax )
		return(0);

	// check directions for exclusion
	dir = m_dxp[0]*(yp-YPnt(1)) - (xp-XPnt(1))*m_dyp[0];
	if (dir*m_dir<0.0)
		return(0);
	dir = m_dxp[1]*(yp-YPnt(2)) - (xp-XPnt(2))*m_dyp[1];
	if (dir*m_dir<0.0)
		return(0);
	dir = m_dxp[2]*(yp-YPnt(0)) - (xp-XPnt(0))*m_dyp[2];
	if (dir*m_dir<0.0)
		return(0);
	return(1);
}

// ------------------------------------------------------------
// calculate zp intercept at a given xp,yp point
double Triangle3DT::ZIntercept(double xp,double yp)
{
	double zp,divisor;
	divisor = m_C;
	if (divisor==0.0) divisor=1.e-6; // avoid division by zero
	zp = -(m_A*xp + m_B*yp + m_D)/divisor;
	return(zp);
}

// ------------------------------------------------------------
void Triangle3DT::SetColor(double r,double g,double b)
{
	int ir,ig,ib;
	// r,g,b (1->100)
	m_red   = r;
	m_green = g;
	m_blue  = b;
	// convert to 0-255 scale
	ir = PCOLOR_TO_WINCOLOR(r);
	ig = PCOLOR_TO_WINCOLOR(g);
	ib = PCOLOR_TO_WINCOLOR(b);
	if (ir<0) ir=0;
	if (ig<0) ig=0;
	if (ib<0) ib=0;
	if (ir>255) ir=255;
	if (ig>255) ig=255;
	if (ib>255) ib=255;
	m_color = RGB(ir,ig,ib); // convert to RGB
}

// ------------------------------------------------------------
void Triangle3DT::Load(unsigned i,unsigned j,unsigned k,Point3DT* pP1,Point3DT* pP2,Point3DT* pP3)
{
 	if (!pP1 || !pP2 || !pP3) return;

	// store input values
 	m_i = i;
 	m_j = j;
 	m_k = k;
 	m_pPnt[0] = pP1;
 	m_pPnt[1] = pP2;
 	m_pPnt[2] = pP3;

	// default min and max projections
 	m_xptmin = m_xptmax = XPnt(0);
 	m_yptmin = m_yptmax = YPnt(0);
 	m_zptmin = m_zptmax = ZPnt(0);

 	// check for max projection
 	if (XPnt(1) > m_xptmax) m_xptmax = XPnt(1);
 	if (YPnt(1) > m_yptmax) m_yptmax = YPnt(1);
 	if (ZPnt(1) > m_zptmax) m_zptmax = ZPnt(1);
 	if (XPnt(2) > m_xptmax) m_xptmax = XPnt(2);
 	if (YPnt(2) > m_yptmax) m_yptmax = YPnt(2);
 	if (ZPnt(2) > m_zptmax) m_zptmax = ZPnt(2);

 	// check for min projection
 	if (XPnt(1) < m_xptmin) m_xptmin = XPnt(1);
 	if (YPnt(1) < m_yptmin) m_yptmin = YPnt(1);
 	if (ZPnt(1) < m_zptmin) m_zptmin = ZPnt(1);
 	if (XPnt(2) < m_xptmin) m_xptmin = XPnt(2);
 	if (YPnt(2) < m_yptmin) m_yptmin = YPnt(2);
 	if (ZPnt(2) < m_zptmin) m_zptmin = ZPnt(2);

 	// calculate point-to-point deltas
 	m_dxp[0] = XPnt(1) - XPnt(0);
 	m_dyp[0] = YPnt(1) - YPnt(0);
 	m_dzp[0] = ZPnt(1) - ZPnt(0);
 	m_dxp[1] = XPnt(2) - XPnt(1);
 	m_dyp[1] = YPnt(2) - YPnt(1);
 	m_dzp[1] = ZPnt(2) - ZPnt(1);
 	m_dxp[2] = XPnt(0) - XPnt(2);
 	m_dyp[2] = YPnt(0) - YPnt(2);
 	m_dzp[2] = ZPnt(0) - ZPnt(2);

 	// calculate plane constants (A*xp + B*yp + C*zp + D = 0)
 	m_A = m_dyp[2]*m_dzp[0] - m_dyp[0]*m_dzp[2];
 	m_B = m_dxp[0]*m_dzp[2] - m_dxp[2]*m_dzp[0];
 	m_C = m_dxp[2]*m_dyp[0] - m_dxp[0]*m_dyp[2];
 	m_D = -(XPnt(0)*m_A + YPnt(0)*m_B + ZPnt(0)*m_C);
	m_area = (m_dxp[0]*m_dyp[1] - m_dxp[1]*m_dyp[0])/2.;
	m_dir = m_area<0.0 ? -1. : 1.;
}

// ------------------------------------------------------------
//     Plot3DT methods
// ------------------------------------------------------------

Plot3DT::Plot3DT()
{
	// zero dynamically allocated pointers
	m_pPoints = 0;
	m_pTriangles = 0;
	m_pHashCounts = 0;
	m_pHashPointers = 0;
	m_nHashItems = 0;

	// set default values
	m_is_color_xyz = 0;
	DefInputsT();
}

Plot3DT::~Plot3DT()
{
	FreeMemory();
}

// ------------------------------------------------------------
void Plot3DT::FreeMemory(void)
{
	unsigned i,count;
	char* pList;
	delete [] m_pPoints;
	delete [] m_pTriangles;
	for (i=0;i<m_nHashItems;i++)
	{
		count = m_pHashCounts[i];
		if (count==0) continue;
		pList = HashListPtr(i);
		delete [] pList;
	} // for i
	delete [] m_pHashCounts;
	delete [] m_pHashPointers;
	m_pPoints = 0;
	m_pTriangles = 0;
	m_pHashCounts = 0;
	m_pHashPointers = 0;
	m_nHashItems = 0;
}

// ------------------------------------------------------------
void Plot3DT::DefInputsT(void)
{
	FreeMemory();	// free memory and clear pointers

	// set default values
	m_nPoints    = 0;
	m_nTriangles = 0;
	m_xpmin  = m_ypmin  = m_zpmin  =  PLOT3D_DBL_MAX; // a very large number
	m_xpmax  = m_ypmax  = m_zpmax  = -PLOT3D_DBL_MAX; // a very small number
	m_maxdxp = m_maxdyp = m_maxdzp = 0.0;
	m_back_color = DEF_BKGND_COLOR;
	m_is_color = 1;
	m_nulines  = 50;
	m_nvlines  = 50;
	m_draw_ulines   = 1;
	m_draw_vlines   = 1;
	m_remove_hidden = 1;
	m_umin = -10;
	m_umax =  10;
	m_vmin = -10;
	m_vmax =  10;
}

// ------------------------------------------------------------
#define	NORMAL_VECTOR_DELTA		(1000)	// delta for partial derivative calculation

// returns: 0=ok, 1=not enough memory
int Plot3DT::CalcPoints()
{
	Point3DT* pPnt;
	double xp,yp,zp;
	double x,y,z;
	double x1,y1,z1;
	double x2,y2,z2;
	double dx1,dy1,dz1;
	double dx2,dy2,dz2;
	double nx,ny,nz, len;
	double du, dv, un, vn;
	double u,v;
	double udelta, vdelta;
	double red=PLOT3D_MAX_COLOR, green=PLOT3D_MAX_COLOR, blue=PLOT3D_MAX_COLOR;
	unsigned i,j;

	// range check to avoid division by zero
	if (m_nulines<2) m_nulines=2;
	if (m_nvlines<2) m_nvlines=2;

	// generate points using parametric equations
	du = (m_umax - m_umin)/double(NORMAL_VECTOR_DELTA);
	dv = (m_vmax - m_vmin)/double(NORMAL_VECTOR_DELTA);

	// user supplied parametric equations
	m_nPoints = m_nulines * m_nvlines;
	m_pPoints = new Point3DT [m_nPoints];
	if (!m_pPoints) return(1);
	udelta = (m_umax - m_umin)/double(m_nulines-1);
	vdelta = (m_vmax - m_vmin)/double(m_nvlines-1);
	for (i=0, u=m_umin; i<m_nulines; i++,u+=udelta)
	{
		for (j=0, v=m_vmin; j<m_nvlines; j++,v+=vdelta)
		{
			// calculate coordinates of point
			pPnt = Point(i,j);
			x  = XParametric(u, v);
			y  = YParametric(u, v);
			z  = ZParametric(u, v);
			if (m_is_color_xyz)
			{
				red   = RedColor  (x,y,z);
				green = GreenColor(x,y,z);
				blue  = BlueColor (x,y,z);
			}
			else
			{
				red   = RedColor  (u,v,0);
				green = GreenColor(u,v,0);
				blue  = BlueColor (u,v,0);
			}
			pPnt->Load(i,j,x,y,z,red,green,blue);
			// calculate normal surface vector at the point
			un = u+du;
			vn = v+dv;
			x1 = XParametric(un,v);
			y1 = YParametric(un,v);
			z1 = ZParametric(un,v);
			x2 = XParametric(u, vn);
			y2 = YParametric(u, vn);
			z2 = ZParametric(u, vn);
			dx1 = x1 - x;
			dy1 = y1 - y;
			dz1 = z1 - z;
			dx2 = x2 - x;
			dy2 = y2 - y;
			dz2 = z2 - z;
			nx =  dy1*dz2 - dy2*dz1;
			ny = -dx1*dz2 + dx2*dz1;
			nz =  dx1*dy2 - dx2*dy1;
			len = sqrt(nx*nx+ny*ny+nz*nz);
			if (len == 0.0)
			{
				nx = ny = nz = 0.;
			}
			else
			{
				// unit normal vector
				nx = nx/len;
				ny = ny/len;
				nz = nz/len;
			}
			// store the unit normal vector
			pPnt->m_xn = nx;
			pPnt->m_yn = ny;
			pPnt->m_zn = nz;
		} // for j
	} // for i

	// project points and calculate max and min projections
	// default min and max to extreme values
	m_xpmin = m_ypmin = m_zpmin =  DBL_MAX; // a very large number
	m_xpmax = m_ypmax = m_zpmax = -DBL_MAX; // a very small (negative) number
	m_xmin  = m_ymin  = m_zmin  =  DBL_MAX; // a very large number
	m_xmax  = m_ymax  = m_zmax  = -DBL_MAX; // a very small (negative) number
 	for (i=0; i<m_nulines; i++)
 	{
 		for (j=0; j<m_nvlines; j++)
 		{
			pPnt = Point(i,j);
			x = pPnt->m_xf;
			y = pPnt->m_yf;
			z = pPnt->m_zf;
			// adjust min and max surface values
			if (x<m_xmin) m_xmin = x;
			if (y<m_ymin) m_ymin = y;
			if (z<m_zmin) m_zmin = z;
			if (x>m_xmax) m_xmax = x;
			if (y>m_ymax) m_ymax = y;
			if (z>m_zmax) m_zmax = z;
			// calculate projection
			xp = xpf(x,y,z);	
			yp = ypf(x,y,z);
			zp = zpf(x,y,z);
 			pPnt->Project(xp,yp,zp);
			// adjust min and max projections
			if (xp<m_xpmin) m_xpmin = xp;
			if (yp<m_ypmin) m_ypmin = yp;
			if (zp<m_zpmin) m_zpmin = zp;
			if (xp>m_xpmax) m_xpmax = xp;
			if (yp>m_ypmax) m_ypmax = yp;
			if (zp>m_zpmax) m_zpmax = zp;
 		} // for j
 	} // for i

	// calculate mid points
	m_xpbmid = (m_xpmin + m_xpmax)/2.;
	m_ypbmid = (m_ypmin + m_ypmax)/2.;
	m_zpbmid = (m_zpmin + m_zpmax)/2.;

	return(0);
}

// ------------------------------------------------------------
#define	PLOT3DT_DISTANCE_TOLERANCE	(0.5)	// percentage of total delta zp
#define	BIN_MULT_FACTOR	(1.2)	// this must be greater than 1.0
#define	EMPTY_HASH	((unsigned)(~0))

// returns: 0=ok, 1=not enough memory
int Plot3DT::CalcTriangles()
{
	Triangle3DT* pTrngl;
	Point3DT* pPnts[3];
	unsigned i,j,k,it,count,nbytes,index;
	unsigned hindex[4];
	double dx,dy,dz;
	double r0,g0,b0,r1,g1,b1,r2,g2,b2,rt,gt,bt;
	char* pList;

 	// allocate memory for triangles
	m_nutriangles = m_nulines-1;
	m_nvtriangles = m_nvlines-1;
 	m_nTriangles = m_nutriangles * m_nvtriangles * 2;
 	m_pTriangles = new Triangle3DT [m_nTriangles];
 	if (!m_pTriangles) return(1);

	// associate points to triangles and calculate maximums
	m_maxdxp = m_maxdyp = m_maxdzp = 0.0;
	for (i=0,it=0; i<m_nutriangles; i++)
	{
		for (j=0; j<m_nvtriangles; j++)
		{
			// lower triangle
			k = 0;
			pTrngl = &m_pTriangles[it++];
			pPnts[0] = Point(i  ,j+1);
			pPnts[1] = Point(i  ,j  );
			pPnts[2] = Point(i+1,j  );
			pTrngl->Load(i,j,k,pPnts[0],pPnts[1],pPnts[2]);
			dx = pTrngl->m_xptmax - pTrngl->m_xptmin;
			dy = pTrngl->m_yptmax - pTrngl->m_yptmin;
			dz = pTrngl->m_zptmax - pTrngl->m_zptmin;
			if (dx > m_maxdxp) m_maxdxp = dx;
			if (dy > m_maxdyp) m_maxdyp = dy;
			if (dz > m_maxdzp) m_maxdzp = dz;

			// upper triangle
			k = 1;
			pTrngl = &m_pTriangles[it++];
			pPnts[0] = Point(i+1,j  );
			pPnts[1] = Point(i+1,j+1);
			pPnts[2] = Point(i  ,j+1);
			pTrngl->Load(i,j,k,pPnts[0],pPnts[1],pPnts[2]);
			dx = pTrngl->m_xptmax - pTrngl->m_xptmin;
			dy = pTrngl->m_yptmax - pTrngl->m_yptmin;
			dz = pTrngl->m_zptmax - pTrngl->m_zptmin;
			if (dx > m_maxdxp) m_maxdxp = dx;
			if (dy > m_maxdyp) m_maxdyp = dy;
			if (dz > m_maxdzp) m_maxdzp = dz;
		} // for j
	} // for i

	// calculate maximum delta of the three directions
	m_zptol = ((m_zpmax-m_zpmin) * PLOT3DT_DISTANCE_TOLERANCE )/100.;

	// calculate hash look-up-table values
	// put triangles into bins based upon location
	dx = m_xpmax - m_xpmin;
	dy = m_ypmax - m_ypmin;
	if (dx==0.0) dx = 1.e-6;	// avoid division by zero
	if (dy==0.0) dx = 1.e-6;	// avoid division by zero
	if (m_maxdxp==0.0) m_maxdxp=1.e-6; // avoid division by zero
	if (m_maxdyp==0.0) m_maxdyp=1.e-6; // avoid division by zero
	m_nxbins = (unsigned)(dx/(m_maxdxp*BIN_MULT_FACTOR) + 1.0);
	m_nybins = (unsigned)(dy/(m_maxdyp*BIN_MULT_FACTOR) + 1.0);
	if (m_nxbins<1) m_nxbins=1;
	if (m_nybins<1) m_nybins=1;
	m_nHashItems = m_nxbins * m_nybins;
	m_dxbin = dx/(double)m_nxbins;
	m_dybin = dy/(double)m_nybins;

	// allocate memory for hash tables
 	m_pHashCounts = new unsigned [m_nHashItems];
	if (!m_pHashCounts) return(1);
	memset(m_pHashCounts,0,sizeof(unsigned)*m_nHashItems);
 	m_pHashPointers = new char * [m_nHashItems];
	if (!m_pHashPointers) return(1);
	memset(m_pHashPointers,0,m_nHashItems*sizeof(char*));

	// determine counts for each hash item
	for (i=0; i<m_nTriangles; i++)
	{
		pTrngl = &m_pTriangles[i];

		// calculate color of triangle from average of points
		r0=pTrngl->RedPnt(0); g0=pTrngl->GreenPnt(0); b0=pTrngl->BluePnt(0);
		r1=pTrngl->RedPnt(1); g1=pTrngl->GreenPnt(1); b1=pTrngl->BluePnt(1);
		r2=pTrngl->RedPnt(2); g2=pTrngl->GreenPnt(2); b2=pTrngl->BluePnt(2);
		// average the three points
		rt = (r0+r1+r2)/3.;
		gt = (g0+g1+g2)/3.;
		bt = (b0+b1+b2)/3.;
		// set the color of the triangle
		pTrngl->SetColor(rt,gt,bt);

		// test four corners of bounding rectangle
		hindex[0] = HashIndex(pTrngl->m_xptmin,pTrngl->m_yptmin);
		hindex[1] = HashIndex(pTrngl->m_xptmax,pTrngl->m_yptmin);
		hindex[2] = HashIndex(pTrngl->m_xptmin,pTrngl->m_yptmax);
		hindex[3] = HashIndex(pTrngl->m_xptmax,pTrngl->m_yptmax);
		// eliminate duplicates
		if (hindex[0] == hindex[1]) hindex[1] = EMPTY_HASH;
		if (hindex[0] == hindex[2]) hindex[2] = EMPTY_HASH;
		if (hindex[0] == hindex[3]) hindex[3] = EMPTY_HASH;
		if (hindex[1] == hindex[2]) hindex[2] = EMPTY_HASH;
		if (hindex[1] == hindex[3]) hindex[3] = EMPTY_HASH;
		if (hindex[2] == hindex[3]) hindex[3] = EMPTY_HASH;
		// increment counts
	 	if (hindex[0] != EMPTY_HASH) m_pHashCounts[hindex[0]]++;
		if (hindex[1] != EMPTY_HASH) m_pHashCounts[hindex[1]]++;
		if (hindex[2] != EMPTY_HASH) m_pHashCounts[hindex[2]]++;
		if (hindex[3] != EMPTY_HASH) m_pHashCounts[hindex[3]]++;
	} // for i

	// allocate memory for hash lists
	for (i=0; i<m_nHashItems; i++)
	{
		count = m_pHashCounts[i];
		if (!count) continue;
		nbytes = count*sizeof(char*);
		pList = new char [nbytes];
		if (!pList) return(1);
		memset(pList,0,nbytes);
		SetHashListPtr(i,pList);
	} // for i

	// generate hash lists
	memset(m_pHashCounts,0,sizeof(unsigned)*m_nHashItems); // clear counts
	for (i=0; i<m_nTriangles; i++)
	{
		pTrngl = &m_pTriangles[i];
		// test four corners of bounding rectangle
		hindex[0] = HashIndex(pTrngl->m_xptmin,pTrngl->m_yptmin);
		hindex[1] = HashIndex(pTrngl->m_xptmax,pTrngl->m_yptmin);
		hindex[2] = HashIndex(pTrngl->m_xptmin,pTrngl->m_yptmax);
		hindex[3] = HashIndex(pTrngl->m_xptmax,pTrngl->m_yptmax);
		// eliminate duplicates
		if (hindex[0] == hindex[1]) hindex[1] = EMPTY_HASH;
		if (hindex[0] == hindex[2]) hindex[2] = EMPTY_HASH;
		if (hindex[0] == hindex[3]) hindex[3] = EMPTY_HASH;
		if (hindex[1] == hindex[2]) hindex[2] = EMPTY_HASH;
		if (hindex[1] == hindex[3]) hindex[3] = EMPTY_HASH;
		if (hindex[2] == hindex[3]) hindex[3] = EMPTY_HASH;
		// store the triangle in the list
	 	if (hindex[0] != EMPTY_HASH)
		{
			index = hindex[0];
			pList = HashListPtr(index);
			count = m_pHashCounts[index];
			SetHashListValuePtr(pList,count,(char*)pTrngl);
			m_pHashCounts[index]++;
		}
	 	if (hindex[1] != EMPTY_HASH)
		{
			index = hindex[1];
			pList = HashListPtr(index);
			count = m_pHashCounts[index];
			SetHashListValuePtr(pList,count,(char*)pTrngl);
			m_pHashCounts[index]++;
		}
	 	if (hindex[2] != EMPTY_HASH)
		{
			index = hindex[2];
			pList = HashListPtr(index);
			count = m_pHashCounts[index];
			SetHashListValuePtr(pList,count,(char*)pTrngl);
			m_pHashCounts[index]++;
		}
	 	if (hindex[3] != EMPTY_HASH)
		{
			index = hindex[3];
			pList = HashListPtr(index);
			count = m_pHashCounts[index];
			SetHashListValuePtr(pList,count,(char*)pTrngl);
			m_pHashCounts[index]++;
		}
	} // for i
	return(0);
}

// ------------------------------------------------------------
// returns: 0=ok
int Plot3DT::InitPlot()
{
	if (InitEquations()) return(1);

	FreeMemory();
	PlotOutputOn();
	CalcTM();		// calculate transformation matrix
	if (CalcPoints()) return(1);	// calculate mesh points
	CalcCorners();	// calculate corners of x,y plate

	// calculate mid points
	m_xpbmid = (m_xpmin + m_xpmax)/2.;
	m_ypbmid = (m_ypmin + m_ypmax)/2.;
	m_zpbmid = (m_zpmin + m_zpmax)/2.;

	// calc plot to video scaling
	double d;
	m_p2v = ((m_scale/100.) * (double)(m_xvmax - m_xvmin)) / (double)(m_xpmax - m_xpmin);
	d     = ((m_scale/100.) * (double)(m_yvmax - m_yvmin)) / (double)(m_ypmax - m_ypmin);
	if (d < m_p2v)
		m_p2v = d;

	// calculate triangles
	// triangles are needed for hidden line removal and exporting
	if (CalcTriangles()) return(1);
	return(0);
}

// ------------------------------------------------------------
int Plot3DT::PlotColorPixels()
{
	CString sPercent;
	long color;
	int xv,yv, percent_complete;

	if (!m_is_color) return(0);
	for (yv=m_yvmin; yv<m_yvmax; yv++)
	{
		// check for abort
		if (IsAborted()) break;

		// draw one scan line
		for (xv=m_xvmin; xv<m_xvmax; xv++)
		{
			color = CalcPixelColor(xv,yv);
			WinSetPixelColor(xv,yv,color);
		} // for xv

		// show percent complete status
		percent_complete = (int)(100.*(yv+1)/m_yvmax);
		sPercent.Format("Colorizing: %u%% complete", percent_complete);
		theApp.SetStatusString(sPercent);
	} // for yv
	return(0);
}

// ------------------------------------------------------------
int Plot3DT::DrawULines()
{
	CString sPercent;
	Point3DT* pPnt1;
	Point3DT* pPnt2;
	unsigned ix,iy,percent_complete;

	if (!m_draw_ulines) return(0);
 	for (ix=0; ix<m_nulines; ix++)
 	{
		if (IsAborted()) break;
		pPnt1 = Point(ix,0);
 		for (iy=1; iy<m_nvlines; iy++)
 		{
			pPnt2 = Point(ix,iy);
			PlotLine(1==iy,pPnt1,pPnt2);
			pPnt1 = pPnt2;
 		} // for iy

		// show percent complete status
		percent_complete = (int)(100.*(ix+1)/m_nulines);
		sPercent.Format("Draw First Contour Lines: %u%% complete", percent_complete);
		theApp.SetStatusString(sPercent);
 	} // for ix
	return(0);
}

// ------------------------------------------------------------
int Plot3DT::DrawVLines()
{
	CString sPercent;
	Point3DT* pPnt1;
	Point3DT* pPnt2;
	unsigned ix,iy,percent_complete;

	if (!m_draw_vlines) return(0);
 	for (iy=0; iy<m_nvlines; iy++)
 	{
		if (IsAborted()) break;
		pPnt1 = Point(0,iy);
 		for (ix=1; ix<m_nulines; ix++)
 		{
			pPnt2 = Point(ix,iy);
			PlotLine(1==ix,pPnt1,pPnt2);
			pPnt1 = pPnt2;
 		} // for ix

		// show percent complete status
		percent_complete = (int)(100.*(iy+1)/m_nvlines);
		sPercent.Format("Draw Second Contour Lines: %u%% complete", percent_complete);
		theApp.SetStatusString(sPercent);
 	} // for iy
	return(0);
}

// ------------------------------------------------------------
void Plot3DT::ResetT(void)
{
	InitBase();
	FreeMemory();
	DefInputsT();
}

// ------------------------------------------------------------
void Plot3DT::GetPOVCamera(double* xc,double* yc,double* zc)
{
	double xdiam,ydiam,diam;
	double zp, scale;
	xdiam = fabs(m_xpmax-m_xpmin);
	ydiam = fabs(m_ypmax-m_ypmin);
	diam  = max(xdiam,ydiam);
	scale = fabs(m_scale);
	if (scale<1) scale=1;
	zp  = (diam/(scale/100.)) + m_zpbmid;
	*xc = xof(m_xpbmid,m_ypbmid,zp);
	*yc = yof(m_xpbmid,m_ypbmid,zp);
	*zc = zof(m_xpbmid,m_ypbmid,zp);
}

// ------------------------------------------------------------
int  Plot3DT::PlotLine(int FirstLine,Point3DT* pPnt1,Point3DT* pPnt2)
{
	double	xp1,yp1,zp1,xp2,yp2,zp2;
	double	dxp,dyp,dzp,ddv;
	int is_vis2, is_vis1;
	int xv1,yv1,xv2,yv2;
	unsigned i,dxv,dyv,dv;

  	xp1 = pPnt1->m_xp;
  	yp1 = pPnt1->m_yp;
  	zp1 = pPnt1->m_zp;

  	xp2 = pPnt2->m_xp;
  	yp2 = pPnt2->m_yp;
  	zp2 = pPnt2->m_zp;

  	xv1 = xvf(xp1);
  	yv1 = yvf(yp1);
  	xv2 = xvf(xp2);
  	yv2 = yvf(yp2);

	dxv = abs(xv2-xv1);
	dyv = abs(yv2-yv1);
	dv  = max(dxv,dyv);
	if (dv<2) dv = 2;
	ddv = double(dv-1);
	dxp = (xp2-xp1)/ddv;
	dyp = (yp2-yp1)/ddv;
	dzp = (zp2-zp1)/ddv;

	// do a move for first line
  	if (FirstLine)
  		FilterMoveTo(xv1,yv1);

  	is_vis1 = IsPointVisible(pPnt1);
	for (i=0; i<dv; i++)
	{
		// calculate point along line segment
		xp2 = xp1 + dxp;
		yp2 = yp1 + dyp;
		zp2 = zp1 + dzp;
	  	xv2 = xvf(xp2);
	  	yv2 = yvf(yp2);

	  	is_vis2 = IsPointVisible(xp2,yp2,zp2);
	
	 	if (is_vis1 && is_vis2)
	 		FilterDrawTo(xv2,yv2);
	 	else
	 		FilterMoveTo(xv2,yv2);

		// update for next pass
		xp1 = xp2;
		yp1 = yp2;
		zp1 = zp2;
		is_vis1 = is_vis2;
	} // for i

	return(0);
}

// ------------------------------------------------------------
long Plot3DT::CalcPixelColor(int xv,int yv)
{
	long color;
	double xp,yp,zp,zpmax=-PLOT3D_DBL_MAX;
	Triangle3DT* pTri;
	Triangle3DT* pTriMax=0;
	unsigned nt,index,count;
	char* pList;
	char* pPtr;

	// convert to plot space
	xp = xv2xp(xv);
	yp = yv2yp(yv);

	// search triangles
	index = HashIndex(xp,yp);
	count = m_pHashCounts[index];
	pList = HashListPtr(index);
	for (nt=0; nt<count; nt++)
	{
		pPtr = HashListValuePtr(pList,nt);
		pTri = (Triangle3DT*)pPtr; // get pointer to triangle

		// check if point is inscribed in triangle
		if (!pTri->IsInscribed(xp,yp)) continue;

		// find closest point
		zp = pTri->ZIntercept(xp,yp);
		if (zp<=zpmax) continue;

		// point is closer
		zpmax = zp;
		pTriMax = pTri;

	} // for nt
	if (pTriMax)
		color = pTriMax->GetColor();
	else
		color = m_back_color;
	return(color);
}

// ------------------------------------------------------------
int Plot3DT::IsPointVisible(double xp,double yp,double zp)
{
	Triangle3DT* pTri;
	unsigned nt;

	if (!m_remove_hidden) return(1);

	// search triangles
	unsigned index = HashIndex(xp,yp);
	unsigned count = m_pHashCounts[index];
	char* pList = HashListPtr(index);
	char* pPtr;
	for (nt=0; nt<count; nt++)
	{
		pPtr = HashListValuePtr(pList,nt);
		pTri = (Triangle3DT*)pPtr; // get pointer to triangle

		// check if point is inscribed in triangle
		if (!pTri->IsInscribed(xp,yp)) continue;

		// check zp intercept to see if it is in front or behind
		if ((pTri->ZIntercept(xp,yp)-m_zptol)<zp) continue;

		// triangle is in front; it's hidden
		return(0);
	} // for nt
	return(1);
}

// ------------------------------------------------------------
int Plot3DT::IsPointVisible(Point3DT* pPnt)
{
	if (pPnt->m_is_visible != -1)
		return(pPnt->m_is_visible);
	pPnt->m_is_visible = IsPointVisible(pPnt->m_xp,pPnt->m_yp,pPnt->m_zp);
	return(pPnt->m_is_visible);
}

// ------------------------------------------------------------
Point3DT* Plot3DT::Point(unsigned i,unsigned j)
{
	unsigned index = (i*m_nvlines) + (j);
	return(&m_pPoints[index]);
}

// ------------------------------------------------------------
unsigned Plot3DT::HashIndex(double xp,double yp)
{
	unsigned index;
	int	ix, iy;
	if (xp<m_xpmin)
		xp = m_xpmin;
	if (yp<m_ypmin)
		yp = m_ypmin;
	if (xp>m_xpmax)
		xp = m_xpmax;
	if (yp>m_ypmax)
		yp = m_ypmax;
	ix = (int)((xp-m_xpmin)/m_dxbin);
	iy = (int)((yp-m_ypmin)/m_dybin);
	index = (iy*m_nxbins) + ix;
	if (index>=m_nHashItems)
		index = m_nHashItems-1;
	return(index);
}

// ------------------------------------------------------------
char* Plot3DT::HashListPtr(unsigned index)
{
	char* pList;
	if (index>=m_nHashItems)
		return(NULL);
	pList = m_pHashPointers[index];
	return(pList);
}

// ------------------------------------------------------------
void Plot3DT::SetHashListPtr(unsigned index,char *pPtr)
{
	if (index>=m_nHashItems)
		return;
	m_pHashPointers[index] = pPtr;
}

// ------------------------------------------------------------
char* Plot3DT::HashListValuePtr(char* pList,unsigned value)
{
	char** pPtrPtr = (char**)pList;
	pList = pPtrPtr[value];
	return(pList);
}

// ------------------------------------------------------------
void Plot3DT::SetHashListValuePtr(char* pList,unsigned value,char *pPtr)
{
	char** pPtrPtr = (char**)pList;
	pPtrPtr[value] = pPtr;
}

// ------------------------------------------------------------
// export file to Persistence-of-Vision
int Plot3DT::ExportPOV(CString povFilePath,CString sPovObjectName,int is_smooth,int is_color)
{
	CString sPovTextureName = "Aluminum";
	Triangle3DT* pTri;
	Point3DT* pP1;
	Point3DT* pP2;
	Point3DT* pP3;
	double xo,yo,zo; // object centroid
	double xc,yc,zc; // camera coordinate
	double xsky,ysky,zsky; // unit vector point to sky
	double xat,yat,zat; // vector pointing from camera to object centroid
	double xright,yright,zright; // unit vector pointing to right side of view
	double red_bg, green_bg, blue_bg; // color components
	double xlight[3], ylight[3], zlight[3]; // coordinates of light sources
	double at_dist;
	unsigned i,nt;
	long bgcolor;
	FILE* fhpov=0;

	if (!IsRendered()) return(0);

	// open file for writing
	fhpov = fopen((LPCTSTR)povFilePath,"w");
	if (fhpov == NULL) return(1);

	// calculate background color values
	bgcolor  = GetBackColor();
	red_bg   = (double)GetRValue(bgcolor)/255.;
	green_bg = (double)GetGValue(bgcolor)/255.;
	blue_bg  = (double)GetBValue(bgcolor)/255.;

	// calculate camera and object coordinates
	GetCenter(&xo,&yo,&zo); // get center of object
	GetPOVCamera(&xc,&yc,&zc);
	zc = -zc; // POV is a lefty
	zo = -zo; // POV is a lefty

	// calclulate sky unit vector
	xsky = xof(0., 1., 0.) - xof(0., 0., 0.);
	ysky = yof(0., 1., 0.) - yof(0., 0., 0.);
	zsky = zof(0., 1., 0.) - zof(0., 0., 0.);
	zsky = -zsky; // POV uses a left-hand coordinate system

	// calculate right unit vector
	xright = xof(1., 0., 0.) - xof(0., 0., 0.);
	yright = yof(1., 0., 0.) - yof(0., 0., 0.);
	zright = zof(1., 0., 0.) - zof(0., 0., 0.);
	zright = -zright; // POV is a lefty

	// calculate look_at vector
	xat = xc-xo;
	yat = yc-yo;
	zat = zc-zo;
	at_dist = sqrt(xat*xat+yat*yat+zat*zat); // distance from camera to object centroid

	// calculate light source coordinates
	// light#0
	xlight[0] = xo + (xat)*.6 + (xsky*at_dist)*0.6 + (xright*at_dist)*0.6;
	ylight[0] = yo + (yat)*.6 + (ysky*at_dist)*0.6 + (yright*at_dist)*0.6;
	zlight[0] = zo + (zat)*.6 + (zsky*at_dist)*0.6 + (zright*at_dist)*0.6;
	// light#1
	xlight[1] = xo + (xat)*.6 + (xsky*at_dist)*0.6 - (xright*at_dist)*0.6;
	ylight[1] = yo + (yat)*.6 + (ysky*at_dist)*0.6 - (yright*at_dist)*0.6;
	zlight[1] = zo + (zat)*.6 + (zsky*at_dist)*0.6 - (zright*at_dist)*0.6;
	// light#2
	xlight[2] = xo + (xat)*.9 - (xsky*at_dist)*0.6 + (xright*at_dist)*.2;
	ylight[2] = yo + (yat)*.9 - (ysky*at_dist)*0.6 + (yright*at_dist)*.2;
	zlight[2] = zo + (zat)*.9 - (zsky*at_dist)*0.6 + (zright*at_dist)*.2;

	// output POV header commands
	fprintf(fhpov,"// POV Image File Generated from SurfX3D\n");
	fprintf(fhpov,"//   Surface: %s\n\n", (LPCTSTR)sPovObjectName);
	fprintf(fhpov,"#include \"colors.inc\"\n");
	fprintf(fhpov,"#include \"textures.inc\"\n\n");
	fprintf(fhpov,"camera {\n");                 
	fprintf(fhpov,"  location <%.6lf, %.6lf, %.6lf>\n",xc,yc,zc); 
	fprintf(fhpov,"  // orthographic\n"); 
	fprintf(fhpov,"  sky <%.6lf, %.6lf, %.6lf>\n",xsky,ysky,zsky); 
	fprintf(fhpov,"  look_at <%.6lf, %.6lf, %.6lf>\n",xo,yo,zo);     
	fprintf(fhpov,"}\n");                        
	fprintf(fhpov,"light_source { <%.6lf, %.6lf, %.6lf> colour White }\n",xlight[0],ylight[0],zlight[0]);
	fprintf(fhpov,"light_source { <%.6lf, %.6lf, %.6lf> colour White }\n",xlight[1],ylight[1],zlight[1]);
	fprintf(fhpov,"light_source { <%.6lf, %.6lf, %.6lf> colour White }\n",xlight[2],ylight[2],zlight[2]);
	fprintf(fhpov,"background { color rgb < %.6lf, %.6lf, %.6lf > }\n", red_bg, green_bg, blue_bg);
	fprintf(fhpov,"\n");

	// loop through each triangle
	nt = TriangleCount();
	fprintf(fhpov,"#declare SURFACE1 = union { // define the object\n");
	for (i=0; i<nt; i++)
	{
		pTri = GetTriangle(i);
		pP1  = pTri->m_pPnt[0];
		pP2  = pTri->m_pPnt[1];
		pP3  = pTri->m_pPnt[2];
		if (is_smooth)
		{
			fprintf(fhpov," smooth_triangle {<%.6lf,%.6lf,%.6lf>,<%.6lf,%.6lf,%.6lf>,<%.6lf,%.6lf,%.6lf>,<%.6lf,%.6lf,%.6lf>,<%.6lf,%.6lf,%.6lf>,<%.6lf,%.6lf,%.6lf>",
				pP1->m_xf,pP1->m_yf,-pP1->m_zf, -pP1->m_xn,-pP1->m_yn,pP1->m_zn,
				pP2->m_xf,pP2->m_yf,-pP2->m_zf, -pP2->m_xn,-pP2->m_yn,pP2->m_zn,
				pP3->m_xf,pP3->m_yf,-pP3->m_zf, -pP3->m_xn,-pP3->m_yn,pP3->m_zn);
		}
		else
		{
			fprintf(fhpov," triangle {<%.6lf,%.6lf,%.6lf>,<%.6lf,%.6lf,%.6lf>,<%.6lf,%.6lf,%.6lf>",
				pP1->m_xf,pP1->m_yf,-pP1->m_zf,
				pP2->m_xf,pP2->m_yf,-pP2->m_zf,
				pP3->m_xf,pP3->m_yf,-pP3->m_zf);
		}

		if (is_color)
		{
			fprintf(fhpov," pigment {color rgb <%.6lf,%.6lf,%.6lf>}}\n",
				pTri->RedTri()  /(double)PLOT3D_MAX_COLOR,	// 0-1
				pTri->GreenTri()/(double)PLOT3D_MAX_COLOR,	// 0-1
				pTri->BlueTri() /(double)PLOT3D_MAX_COLOR);	// 0-1
		}
		else
		{
			fprintf(fhpov,"}\n");
		}

	} // for i
	fprintf(fhpov," texture { %s } }\n", (LPCTSTR)sPovTextureName);

	// usage
	fprintf(fhpov,"\nobject { SURFACE1 } // use the object\n");

	// cleanup
	fclose(fhpov);
	return(0);
}

// ------------------------------------------------------------
// export file to Virtual-Reality-Modeling-Language (VRML) file format
int Plot3DT::ExportVRML(CString sVrmlFilePath,CString sVrmlSurfaceName,int isColor,int isBack,int drawLine1,int drawLine2)
{
	Point3DT* pPnt;
	float red=0.,green=0.,blue=0.;
	int i1,i2,i3,i4;
	unsigned i,j,np;
	FILE* fhvrml=0;

	if (!IsRendered()) return(0);

	// open file for writing
	fhvrml = fopen((LPCTSTR)sVrmlFilePath,"w");
	if (fhvrml == NULL) return(1);

	// ----------------
	// VRML File Header
	// ----------------
	fprintf(fhvrml,"#VRML V2.0 utf8\n");
	fprintf(fhvrml,"# Generated by SurfX3D\n");
	fprintf(fhvrml,"\n");
	fprintf(fhvrml,"WorldInfo { title \"%s\" }\n", sVrmlSurfaceName);
	fprintf(fhvrml,"\n");

	// background color
	if (isBack)
	{
		// Back Color Components 0-1
		red   = (float)(GetRValue(m_back_color)/255.);
		green = (float)(GetGValue(m_back_color)/255.);
		blue  = (float)(GetBValue(m_back_color)/255.);

		fprintf(fhvrml,"Background {\n");
		fprintf(fhvrml,"  skyColor    [ %.4f %.4f %.4f ]", red, green, blue);
		fprintf(fhvrml,"  groundColor [ %.4f %.4f %.4f ]", red, green, blue);
		fprintf(fhvrml,"}");
		fprintf(fhvrml,"\n");
	}

	// ---------------
	// PointSet Shape
	// ---------------
	fprintf(fhvrml,"DEF POINTS1 Shape {\n");
	fprintf(fhvrml," geometry PointSet {\n");
	fprintf(fhvrml,"   coord DEF COORD1 Coordinate {\n");
	fprintf(fhvrml,"     point [\n");
	for (j=np=0; j<m_nvlines; j++)
 	{
	 	for (i=0; i<m_nulines; i++)
 		{
			pPnt = Point(i,j);
			fprintf(fhvrml,"      %10.6lf %10.6lf %10.6lf # %u\n", pPnt->m_xf, pPnt->m_yf, pPnt->m_zf, np++);
		} // for j
	} // for i
	fprintf(fhvrml,"     ] # point\n");
	fprintf(fhvrml,"   }   # coord\n");
	fprintf(fhvrml," }     # geometry\n");
	fprintf(fhvrml,"}      # Shape\n");


	// -------------------
	// Point Surface Shape
	// -------------------
	if (isColor)
	{
		fprintf(fhvrml,"DEF SURFACE1 Shape {\n");
		fprintf(fhvrml," geometry IndexedFaceSet {\n");
		fprintf(fhvrml,"   solid FALSE\n");
		fprintf(fhvrml,"   coord USE COORD1\n");
		fprintf(fhvrml,"   color Color\n");
		fprintf(fhvrml,"   {\n");
		fprintf(fhvrml,"     color [\n");
		for (j=np=0; j<m_nvlines; j++)
	 	{
		 	for (i=0; i<m_nulines; i++)
	 		{
				pPnt = Point(i,j);
				fprintf(fhvrml,"      %10.6lf %10.6lf %10.6lf # %u\n", pPnt->m_red/100., pPnt->m_green/100., pPnt->m_blue/100., np++);
			} // for j
		} // for i
		fprintf(fhvrml,"     ]\n");
		fprintf(fhvrml,"   }\n");
	
		// face set normal vectors
		fprintf(fhvrml,"   normal Normal {\n");
		fprintf(fhvrml,"     vector [\n");
		for (j=np=0; j<m_nvlines; j++)
		{
			for (i=0; i<m_nulines; i++)
			{
				pPnt = Point(i,j);
				fprintf(fhvrml,"      %10.6lf %10.6lf %10.6lf # %u\n", pPnt->m_xn, pPnt->m_yn, pPnt->m_zn, np++);
			} // for j
		} // for i
		fprintf(fhvrml,"     ]\n");
		fprintf(fhvrml,"   }\n");
	
		// face set coordinate indexes
		fprintf(fhvrml,"   coordIndex [\n");
		for (j=0; j<m_nvlines-1; j++)
	 	{
		 	for (i=0; i<m_nulines-1; i++)
	 		{
				i1 = j*m_nulines+i;
				i2 = i1 + 1;
				i3 = i1 + m_nulines;
				i4 = i3 + 1;
				fprintf(fhvrml,"      %6u %6u %6u -1\n", i1,i2,i3);
				fprintf(fhvrml,"      %6u %6u %6u -1\n", i2,i4,i3);
			} // for j
		} // for i
		fprintf(fhvrml,"   ]\n");
	
		// trailer
		fprintf(fhvrml," }\n");
		fprintf(fhvrml,"}\n");
	} // m_is_color

	// -------------------------
	// Line Color Components 0-1
	// -------------------------
	red   = (float)(GetRValue(m_line_color)/255.);
	green = (float)(GetGValue(m_line_color)/255.);
	blue  = (float)(GetBValue(m_line_color)/255.);

	// ---------------
	// Line Set1 Shape
	// ---------------
	if (drawLine1)
	{
		fprintf(fhvrml,"DEF LINESET1 Shape {\n");
		fprintf(fhvrml," appearance Appearance { material Material { emissiveColor %.4f %.4f %.4f } }\n", red, green, blue);
		fprintf(fhvrml," geometry IndexedLineSet {\n");
		fprintf(fhvrml,"   coord USE COORD1\n");
		fprintf(fhvrml,"   coordIndex [\n");
		for (j=0; j<m_nvlines; j++)
	 	{
		 	for (i=0; i<m_nulines-1; i++)
	 		{
				i1 = j*m_nulines+i;
				i2 = i1 + 1;
				fprintf(fhvrml,"      %6u %6u -1\n", i1,i2);
			} // for j
		} // for i
		fprintf(fhvrml,"   ]\n");
		fprintf(fhvrml," }\n");
		fprintf(fhvrml,"}\n");
	}

	// ---------------
	// Line Set2 Shape
	// ---------------
	if (drawLine2)
	{
		fprintf(fhvrml,"DEF LINESET2 Shape {\n");
		fprintf(fhvrml," appearance Appearance { material Material { emissiveColor %.4f %.4f %.4f } }\n", red, green, blue);
		fprintf(fhvrml," geometry IndexedLineSet {\n");
		fprintf(fhvrml,"   coord USE COORD1\n");
		fprintf(fhvrml,"   coordIndex [\n");
	 	for (i=0; i<m_nulines; i++)
	 	{
			for (j=0; j<m_nvlines-1; j++)
	 		{
				i1 = j*m_nulines+i;
				i2 = (j+1)*m_nulines+i;
				fprintf(fhvrml,"      %6u %6u -1\n", i1,i2);
			} // for j
		} // for i
		fprintf(fhvrml,"   ]\n");
		fprintf(fhvrml," }\n");
		fprintf(fhvrml,"}\n");
	}

	// -------
	// cleanup
	// -------
	fclose(fhvrml);
	return(0);
}

// ------------------------------------------------------------
// export file to Drawing Exchange File (DXF) format
int Plot3DT::ExportDXF(CString sDxfFilePath,CString sDxfSurfaceName)
{
	Point3DT* pPnt;
	unsigned i,j;
	FILE* fhdxf=0;

	if (!IsRendered()) return(0);

	// open file for writing
	fhdxf = fopen((LPCTSTR)sDxfFilePath,"w");
	if (fhdxf == NULL) return(1);

	// ---------------
	// DXF File Header
	// ---------------
	fprintf(fhdxf,"999\n");
	fprintf(fhdxf,"DXF created from SurfX3D: %s\n", sDxfSurfaceName);
	fprintf(fhdxf,"0\n");
	fprintf(fhdxf,"SECTION\n");
	fprintf(fhdxf,"2\n");
	fprintf(fhdxf,"HEADER\n");
	fprintf(fhdxf,"0\n");
	fprintf(fhdxf,"ENDSEC\n");
	fprintf(fhdxf,"0\n");
	fprintf(fhdxf,"SECTION\n");
	fprintf(fhdxf,"2\n");
	fprintf(fhdxf,"TABLES\n");
	fprintf(fhdxf,"0\n");
	fprintf(fhdxf,"ENDSEC\n");
	fprintf(fhdxf,"0\n");
	fprintf(fhdxf,"SECTION\n");
	fprintf(fhdxf,"2\n");
	fprintf(fhdxf,"BLOCKS\n");
	fprintf(fhdxf,"0\n");
	fprintf(fhdxf,"ENDSEC\n");
	fprintf(fhdxf,"0\n");
	fprintf(fhdxf,"SECTION\n");
	fprintf(fhdxf,"2\n");
	fprintf(fhdxf,"ENTITIES\n");

	// -------------
	// Surface Shape
	// -------------
	for (j=0; j<m_nvlines-1; j++)
	{
	 	for (i=0; i<m_nulines-1; i++)
		{
			fprintf(fhdxf,"0\n");
			fprintf(fhdxf,"3DFACE\n");
			fprintf(fhdxf,"8\n");
			fprintf(fhdxf,"0\n");
			fprintf(fhdxf,"70\n");
			fprintf(fhdxf,"0\n");

			pPnt = Point(i,j);
			fprintf(fhdxf,"10\n");
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_xf);
			fprintf(fhdxf,"20\n");
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_yf);
			fprintf(fhdxf,"30\n");
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_zf);

			pPnt = Point(i,j+1);
			fprintf(fhdxf,"11\n");
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_xf);
			fprintf(fhdxf,"21\n");	            
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_yf);
			fprintf(fhdxf,"31\n");	            
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_zf);

			pPnt = Point(i+1,j+1);
			fprintf(fhdxf,"12\n");
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_xf);
			fprintf(fhdxf,"22\n");	            
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_yf);
			fprintf(fhdxf,"32\n");	            
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_zf);

			pPnt = Point(i+1,j);
			fprintf(fhdxf,"13\n");
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_xf);
			fprintf(fhdxf,"23\n");	            
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_yf);
			fprintf(fhdxf,"33\n");	            
			fprintf(fhdxf,"%10.6lf\n",pPnt->m_zf);
		} // for j
	} // for i

	// -------
	// cleanup
	// -------
	fprintf(fhdxf,"0\n");
	fprintf(fhdxf,"ENDSEC\n");
	fprintf(fhdxf,"0\n");
	fprintf(fhdxf,"EOF\n");

	fclose(fhdxf);
	return(0);
}

//	<><><><><><><><><><><><><>  Plot3DT.cpp  <><><><><><><><><><><><><><> //














